#ifndef CARD_H
#define CARD_H

#include <QBitArray>
#include <QHash>
#include <QList>
#include <QPointer>
#include <QDebug>
#include <QUndoStack>
#include <QFile>
#include <QSharedPointer>

#include <QModelIndex>

namespace QPunchCard {
	class Column; class Card; class Deck; class DeckIndex;
};

#include "format.h"
//#include "codec.h"

namespace QPunchCard {

/**
 * A Column is a 12 bit storage...
 **/
class Column : public QBitArray {
public:
	Column() : QBitArray(13) {}

	Column(int value) : QBitArray(13) {
		// TODO: range check, throw exception if bad
		for(int s=0; s<12; s++) {
			setBit(s, (1 << s) & value);
			//setBit(s, (value >> s) & 0x01);
		}
	}


	operator int() const {
		int ret=0;
		for(int s=0; s<12; s++)
			ret |= testBit(s) << s;
		return ret;
	}
};

/**
 * This is a punch card.
 **/
class Card : public QObject {
	Q_OBJECT // must be qobject for signals/slots and QPointer usage

public:
	Card() {};
	QHash<QString,QString> properties;

	Column column[80];
	Column& get(int x) {
		if(x < 0 || x > 80)
			throw "Out of range exception";
		return column[x];
	}
	const Column& get(int x) const {
		if(x < 0 || x > 80)
			throw "Out of range exception";
		return column[x];
	}
	Column& operator[](int x) { return get(x); }
	const Column& operator[](int x) const { return column[x]; }
	int length() const { return 80; } // aja...

	/*Card& operator=(const Card& other) {
		// copy assignment, etc.
		for(int x=0; x<80; x++) column[x] = other[x];
		return *this;
	}*/

	~Card() {};

signals:
	// auf das level will echt keiner runter!
	void dataChanged(int left_index, int right_index);
	// wenn signal: klasse muss QObject sein =>
	//              kein copy constructor mglich =>
	//              File = QVector<Card*>
};


/**
 * File uses the strategy pattern to switch between different
 * I/O formats (using the FileFormat class).
 **/
class Deck : public QObject, public QList< QPointer<Card> > {
	Q_OBJECT

	QSharedPointer<const FileFormat> created_format;
	void init();

public:
	typedef QList< QPointer<Card> >::iterator iterator;
	QPointer<QUndoStack> undo;

	// Creation and I/O handling ===================
	/// Create empty
	Deck() : QList< QPointer<Card> >() { init(); }
	~Deck() {};
	/// Create from file/stream/etc., that is, calls format->read()
	Deck(const FileFormat* format) : QList< QPointer<Card> >(), created_format(format) { init(); }
	/// Save with same device/format as created
	bool save(QFile& file) { return save(created_format.data(), file); }
	/// Save
	bool save(const FileFormat* format, QFile& file);
	bool read(QFile& file) { return read(created_format.data(), file); }
	bool read(const FileFormat* format, QFile& file);

	bool canSave() const { return created_format; }
	void setFormat(const FileFormat* format);
	//File* getFile() { if(created_format && created_format->file) return created_format->file; }

	bool isModified() const { return !undo->isClean(); }

	// index handling ======================
	inline bool isValid(int i);
	inline DeckIndex createIndex(int i);

	// Bearbeitungsfunktionen
	bool insert(DeckIndex after); // neue Karte hinzufuegen
	bool insertTimes(DeckIndex after, int times = 1);
	bool move(DeckIndex from, DeckIndex to);
	bool erase(DeckIndex from, DeckIndex to);

	// Quick & dirty
	inline void emitChanged(DeckIndex lower, DeckIndex upper);
signals:
	void cardCountChanged(int newCardCount); // wer braucht das?

	// das hier ist das FAVORISIERTE FORMAT:
	void contentsChanged(DeckIndex lowerCard, DeckIndex upperCard);


	void modified(bool modified);
	//void redoAvailable(bool available);
	//void undoAvailable(bool available);
	//void undoCommandAdded();

public slots:
	//void redo() {}
	//void undo() {}
	//void setModified(bool modified) {}
};

// Java-style iterator
typedef QListIterator< QPointer<Card> > DeckIterator;


/**
 * CardIndex feels like an int (normalized)
 **/
class DeckIndex {
private:
	QPointer<Deck> deck;
	int i;
public:
	// behaves like int, copy constructor and operator= will be created automatically
	DeckIndex(int i = 0) : i(i) {}
	operator int() const { return normalized(); }
	int asInt() const { return normalized(); }

	DeckIndex(Deck* d, int i) : deck(d), i(i) { if(!deck) qDebug() << "NULL DECK!"; }
	bool isValid() const { return deck ? (i >= 0 && i < deck->count()) : false; }

	// "normalisierung" = in Deck Bounds
	int normalized() const { if(!deck) return i; if(i < 0) return 0; else if(i >= deck->count()) return deck->count()-1; else return i; }
	void normalize() { i = normalized(); }
	bool canNormalize() const { return deck; }

	// special positions
	bool isTop() const { return i < 0; }
	bool isEnd() const { return i >= deck->count(); }
	bool isFirst() const { return i == 0; }
	bool isLast() const { return i == deck->count(); }

	const QPointer<Deck> getDeck() const { return deck; }
	void setDeck(Deck* deck) { this->deck = deck; }
	bool assertDeck(Deck* fallback) { if(!deck) deck=fallback; return deck == fallback; }
	bool hasDeck() const { return deck; }

	// ein paar abkuerzungen
	/// @returns null wenn nicht normalisierbar (also letztlich kein Deck da ist)
	QPointer<Card> getCard() const { return canNormalize() ? deck->at(normalized()) : NULL; }

	DeckIndex& operator++() { i++; return *this; }
	DeckIndex& operator+=(int x) { i+= x; return *this; }
};

/****************************************************************************
  debug functions
 ***************************************************************************/

 QDebug operator<<(QDebug dbg, const Column &c);
 QDebug operator<<(QDebug dbg, const Card &c);
 QDebug operator<<(QDebug dbg, const Deck &c);
 QDebug operator<<(QDebug dbg, const DeckIndex &c);


/****************************************************************************
  inline functions
 ***************************************************************************/

inline bool Deck::isValid(int i) { return createIndex(i).isValid(); }
inline DeckIndex Deck::createIndex(int i) { return DeckIndex(this, i); }
inline void Deck::emitChanged(DeckIndex lower, DeckIndex upper) {
	emit contentsChanged(lower, upper);
}

}; // Namespace
#endif // CARD_H
